@@ -65,7 +65,6 @@ row_undo_ins_remove_clust_rec(
 /*==========================*/
 	undo_node_t*	node)	/*!< in: undo node */
 {
-	btr_cur_t*	btr_cur;
 	ibool		success;
 	dberr_t		err;
 	ulint		n_tries	= 0;
@@ -73,15 +72,27 @@ row_undo_ins_remove_clust_rec(
 	dict_index_t*	index	= node->pcur.btr_cur.index;
 	bool		online;
 
-	ut_ad(dict_index_is_clust(index));
+	ut_ad(index->is_primary());
 	ut_ad(node->trx->in_rollback);
 
 	mtr.start();
 	if (index->table->is_temporary()) {
 		ut_ad(node->rec_type == TRX_UNDO_INSERT_REC);
 		mtr.set_log_mode(MTR_LOG_NO_REDO);
+		ut_ad(!dict_index_is_online_ddl(index));
+		ut_ad(index->table->id >= DICT_HDR_FIRST_ID);
+		online = false;
 	} else {
 		index->set_modified(mtr);
+		online = dict_index_is_online_ddl(index);
+		if (online) {
+			ut_ad(node->rec_type == TRX_UNDO_INSERT_REC);
+			ut_ad(node->trx->dict_operation_lock_mode
+			      != RW_X_LATCH);
+			ut_ad(node->table->id != DICT_INDEXES_ID);
+			ut_ad(node->table->id != DICT_COLUMNS_ID);
+			mtr_s_lock(dict_index_get_lock(index), &mtr);
+		}
 	}
 
 	/* This is similar to row_undo_mod_clust(). The DDL thread may
@@ -90,105 +101,69 @@ row_undo_ins_remove_clust_rec(
 	purged. However, we can log the removal out of sync with the
 	B-tree modification. */
 
-	online = dict_index_is_online_ddl(index);
-	if (online) {
-		ut_ad(node->trx->dict_operation_lock_mode
-		      != RW_X_LATCH);
-		ut_ad(node->table->id != DICT_INDEXES_ID);
-		mtr_s_lock_index(index, &mtr);
-	}
-
 	success = btr_pcur_restore_position(
 		online
 		? BTR_MODIFY_LEAF | BTR_ALREADY_S_LATCHED
-		: BTR_MODIFY_LEAF, &node->pcur, &mtr);
+		: (node->rec_type == TRX_UNDO_INSERT_METADATA)
+		? BTR_MODIFY_TREE : BTR_MODIFY_LEAF, &node->pcur, &mtr);
 	ut_a(success);
 
-	btr_cur = btr_pcur_get_btr_cur(&node->pcur);
+	rec_t* rec = btr_pcur_get_rec(&node->pcur);
 
-	ut_ad(rec_get_trx_id(btr_cur_get_rec(btr_cur), btr_cur->index)
-	      == node->trx->id);
-	ut_ad(!rec_get_deleted_flag(
-		      btr_cur_get_rec(btr_cur),
-		      dict_table_is_comp(btr_cur->index->table)));
+	ut_ad(rec_get_trx_id(rec, index) == node->trx->id);
+	ut_ad(!rec_get_deleted_flag(rec, index->table->not_redundant())
+	      || rec_is_alter_metadata(rec, index->table->not_redundant()));
+	ut_ad(rec_is_metadata(rec, index->table->not_redundant())
+	      == (node->rec_type == TRX_UNDO_INSERT_METADATA));
 
 	if (online && dict_index_is_online_ddl(index)) {
-		const rec_t*	rec	= btr_cur_get_rec(btr_cur);
 		mem_heap_t*	heap	= NULL;
 		const ulint*	offsets	= rec_get_offsets(
 			rec, index, NULL, true, ULINT_UNDEFINED, &heap);
 		row_log_table_delete(rec, index, offsets, NULL);
 		mem_heap_free(heap);
+	} else {
+		switch (node->table->id) {
+		case DICT_INDEXES_ID:
+			ut_ad(!online);
+			ut_ad(node->trx->dict_operation_lock_mode
+			      == RW_X_LATCH);
+			ut_ad(node->rec_type == TRX_UNDO_INSERT_REC);
+
+			dict_drop_index_tree(rec, &node->pcur, &mtr);
+			mtr.commit();
+
+			mtr.start();
+			success = btr_pcur_restore_position(
+				BTR_MODIFY_LEAF, &node->pcur, &mtr);
+			ut_a(success);
+			break;
+		case DICT_COLUMNS_ID:
+			/* This is rolling back an INSERT into SYS_COLUMNS.
+			If it was part of an instant ALTER TABLE operation, we
+			must evict the table definition, so that it can be
+			reloaded after the dictionary operation has been
+			completed. At this point, any corresponding operation
+			to the metadata record will have been rolled back. */
+			ut_ad(!online);
+			ut_ad(node->trx->dict_operation_lock_mode
+			      == RW_X_LATCH);
+			ut_ad(node->rec_type == TRX_UNDO_INSERT_REC);
+			if (rec_get_n_fields_old(rec)
+			    != DICT_NUM_FIELDS__SYS_COLUMNS) {
+				break;
+			}
+			ulint len;
+			const byte* data = rec_get_nth_field_old(
+				rec, DICT_FLD__SYS_COLUMNS__TABLE_ID, &len);
+			if (len != 8) {
+				break;
+			}
+			node->trx->evict_table(mach_read_from_8(data));
+		}
 	}
 
-	switch (node->table->id) {
-	case DICT_INDEXES_ID:
-		ut_ad(!online);
-		ut_ad(node->trx->dict_operation_lock_mode == RW_X_LATCH);
-		ut_ad(node->rec_type == TRX_UNDO_INSERT_REC);
-
-		dict_drop_index_tree(
-			btr_pcur_get_rec(&node->pcur), &node->pcur, node->trx,
-			&mtr);
-
-		mtr.commit();
-
-		mtr.start();
-
-		success = btr_pcur_restore_position(
-			BTR_MODIFY_LEAF, &node->pcur, &mtr);
-		ut_a(success);
-		break;
-	case DICT_COLUMNS_ID:
-		/* This is rolling back an INSERT into SYS_COLUMNS.
-		If it was part of an instant ADD COLUMN operation, we
-		must modify the table definition. At this point, any
-		corresponding operation to the metadata record will have
-		been rolled back. */
-		ut_ad(!online);
-		ut_ad(node->trx->dict_operation_lock_mode == RW_X_LATCH);
-		ut_ad(node->rec_type == TRX_UNDO_INSERT_REC);
-		const rec_t* rec = btr_pcur_get_rec(&node->pcur);
-		if (rec_get_n_fields_old(rec)
-		    != DICT_NUM_FIELDS__SYS_COLUMNS) {
-			break;
-		}
-		ulint len;
-		const byte* data = rec_get_nth_field_old(
-			rec, DICT_FLD__SYS_COLUMNS__TABLE_ID, &len);
-		if (len != 8) {
-			break;
-		}
-		const table_id_t table_id = mach_read_from_8(data);
-		data = rec_get_nth_field_old(rec, DICT_FLD__SYS_COLUMNS__POS,
-					     &len);
-		if (len != 4) {
-			break;
-		}
-		const unsigned pos = mach_read_from_4(data);
-		if (pos == 0 || pos >= (1U << 16)) {
-			break;
-		}
-		dict_table_t* table = dict_table_open_on_id(
-			table_id, true, DICT_TABLE_OP_OPEN_ONLY_IF_CACHED);
-		if (!table) {
-			break;
-		}
-
-		dict_index_t* index = dict_table_get_first_index(table);
-
-		if (index && index->is_instant()
-		    && DATA_N_SYS_COLS + 1 + pos == table->n_cols) {
-			/* This is the rollback of an instant ADD COLUMN.
-			Remove the column from the dictionary cache,
-			but keep the system columns. */
-			table->rollback_instant(pos);
-		}
-
-		dict_table_close(table, true, false);
-	}
-
-	if (btr_cur_optimistic_delete(btr_cur, 0, &mtr)) {
+	if (btr_cur_optimistic_delete(&node->pcur.btr_cur, 0, &mtr)) {
 		err = DB_SUCCESS;
 		goto func_exit;
 	}
@@ -208,7 +183,8 @@ retry:
 			&node->pcur, &mtr);
 	ut_a(success);
 
-	btr_cur_pessimistic_delete(&err, FALSE, btr_cur, 0, true, &mtr);
+	btr_cur_pessimistic_delete(&err, FALSE, &node->pcur.btr_cur, 0, true,
+				   &mtr);
 
 	/* The delete operation may fail if we have little
 	file space left: TODO: easiest to crash the database
@@ -227,29 +203,34 @@ retry:
 	}
 
 func_exit:
-	btr_pcur_commit_specify_mtr(&node->pcur, &mtr);
 	if (err == DB_SUCCESS && node->rec_type == TRX_UNDO_INSERT_METADATA) {
 		/* When rolling back the very first instant ADD COLUMN
 		operation, reset the root page to the basic state. */
 		ut_ad(!index->table->is_temporary());
-		mtr.start();
 		if (page_t* root = btr_root_get(index, &mtr)) {
 			byte* page_type = root + FIL_PAGE_TYPE;
 			ut_ad(mach_read_from_2(page_type)
 			      == FIL_PAGE_TYPE_INSTANT
 			      || mach_read_from_2(page_type)
 			      == FIL_PAGE_INDEX);
-			index->set_modified(mtr);
 			mlog_write_ulint(page_type, FIL_PAGE_INDEX,
 					 MLOG_2BYTES, &mtr);
 			byte* instant = PAGE_INSTANT + PAGE_HEADER + root;
 			mlog_write_ulint(instant,
 					 page_ptr_get_direction(instant + 1),
 					 MLOG_2BYTES, &mtr);
+			rec_t* infimum = page_get_infimum_rec(root);
+			rec_t* supremum = page_get_supremum_rec(root);
+			static const byte str[8 + 8] = "supremuminfimum";
+			if (memcmp(infimum, str + 8, 8)
+			    || memcmp(supremum, str, 8)) {
+				mlog_write_string(infimum, str + 8, 8, &mtr);
+				mlog_write_string(supremum, str, 8, &mtr);
+			}
 		}
-		mtr.commit();
 	}
 
+	btr_pcur_commit_specify_mtr(&node->pcur, &mtr);
 	return(err);
 }
 
@@ -276,10 +257,10 @@ row_undo_ins_remove_sec_low(
 
 	if (modify_leaf) {
 		mode = BTR_MODIFY_LEAF | BTR_ALREADY_S_LATCHED;
-		mtr_s_lock_index(index, &mtr);
+		mtr_s_lock(dict_index_get_lock(index), &mtr);
 	} else {
 		ut_ad(mode == (BTR_MODIFY_TREE | BTR_LATCH_FOR_DELETE));
-		mtr_sx_lock_index(index, &mtr);
+		mtr_sx_lock(dict_index_get_lock(index), &mtr);
 	}
 
 	if (row_log_online_op_try(index, entry, 0)) {
@@ -381,14 +362,10 @@ retry:
 	return(err);
 }
 
-/***********************************************************//**
-Parses the row reference and other info in a fresh insert undo record. */
-static
-void
-row_undo_ins_parse_undo_rec(
-/*========================*/
-	undo_node_t*	node,		/*!< in/out: row undo node */
-	ibool		dict_locked)	/*!< in: TRUE if own dict_sys->mutex */
+/** Parse an insert undo record.
+@param[in,out]	node		row rollback state
+@param[in]	dict_locked	whether the data dictionary cache is locked */
+static bool row_undo_ins_parse_undo_rec(undo_node_t* node, bool dict_locked)
 {
 	dict_index_t*	clust_index;
 	byte*		ptr;
@@ -397,18 +374,28 @@ row_undo_ins_parse_undo_rec(
 	ulint		dummy;
 	bool		dummy_extern;
 
-	ut_ad(node);
+	ut_ad(node->state == UNDO_INSERT_PERSISTENT
+	      || node->state == UNDO_INSERT_TEMPORARY);
+	ut_ad(node->trx->in_rollback);
+	ut_ad(trx_undo_roll_ptr_is_insert(node->roll_ptr));
 
 	ptr = trx_undo_rec_get_pars(node->undo_rec, &node->rec_type, &dummy,
 				    &dummy_extern, &undo_no, &table_id);
 
 	node->update = NULL;
-	node->table = dict_table_open_on_id(
-		table_id, dict_locked, DICT_TABLE_OP_NORMAL);
+	if (node->state == UNDO_INSERT_PERSISTENT) {
+		node->table = dict_table_open_on_id(table_id, dict_locked,
+						    DICT_TABLE_OP_NORMAL);
+	} else if (!dict_locked) {
+		mutex_enter(&dict_sys.mutex);
+		node->table = dict_sys.get_temporary_table(table_id);
+		mutex_exit(&dict_sys.mutex);
+	} else {
+		node->table = dict_sys.get_temporary_table(table_id);
+	}
 
-	/* Skip the UNDO if we can't find the table or the .ibd file. */
-	if (UNIV_UNLIKELY(node->table == NULL)) {
-		return;
+	if (!node->table) {
+		return false;
 	}
 
 	switch (node->rec_type) {
@@ -447,6 +434,7 @@ close_table:
 		connection, instead of doing this rollback. */
 		dict_table_close(node->table, dict_locked, FALSE);
 		node->table = NULL;
+		return false;
 	} else {
 		ut_ad(!node->table->skip_alter_undo);
 		clust_index = dict_table_get_first_index(node->table);
@@ -478,6 +466,8 @@ close_table:
 			goto close_table;
 		}
 	}
+
+	return true;
 }
 
 /***************************************************************//**
@@ -554,18 +544,10 @@ row_undo_ins(
 	que_thr_t*	thr)	/*!< in: query thread */
 {
 	dberr_t	err;
-	ibool	dict_locked;
+	bool dict_locked = node->trx->dict_operation_lock_mode == RW_X_LATCH;
 
-	ut_ad(node->state == UNDO_NODE_INSERT);
-	ut_ad(node->trx->in_rollback);
-	ut_ad(trx_undo_roll_ptr_is_insert(node->roll_ptr));
-
-	dict_locked = node->trx->dict_operation_lock_mode == RW_X_LATCH;
-
-	row_undo_ins_parse_undo_rec(node, dict_locked);
-
-	if (node->table == NULL) {
-		return(DB_SUCCESS);
+	if (!row_undo_ins_parse_undo_rec(node, dict_locked)) {
+		return DB_SUCCESS;
 	}
 
 	/* Iterate over all the indexes and undo the insert.*/
@@ -589,26 +571,19 @@ row_undo_ins(
 			break;
 		}
 
-		/* fall through */
-	case TRX_UNDO_INSERT_METADATA:
 		log_free_check();
 
 		if (node->table->id == DICT_INDEXES_ID) {
-			ut_ad(node->rec_type == TRX_UNDO_INSERT_REC);
-
+			ut_ad(!node->table->is_temporary());
 			if (!dict_locked) {
-				mutex_enter(&dict_sys->mutex);
+				mutex_enter(&dict_sys.mutex);
 			}
-		}
-
-		// FIXME: We need to update the dict_index_t::space and
-		// page number fields too.
-		err = row_undo_ins_remove_clust_rec(node);
-
-		if (node->table->id == DICT_INDEXES_ID
-		    && !dict_locked) {
-
-			mutex_exit(&dict_sys->mutex);
+			err = row_undo_ins_remove_clust_rec(node);
+			if (!dict_locked) {
+				mutex_exit(&dict_sys.mutex);
+			}
+		} else {
+			err = row_undo_ins_remove_clust_rec(node);
 		}
 
 		if (err == DB_SUCCESS && node->table->stat_initialized) {
@@ -621,13 +596,19 @@ row_undo_ins(
 			/* Do not attempt to update statistics when
 			executing ROLLBACK in the InnoDB SQL
 			interpreter, because in that case we would
-			already be holding dict_sys->mutex, which
+			already be holding dict_sys.mutex, which
 			would be acquired when updating statistics. */
 			if (!dict_locked) {
 				dict_stats_update_if_needed(
 					node->table, node->trx->mysql_thd);
 			}
 		}
+		break;
+
+	case TRX_UNDO_INSERT_METADATA:
+		log_free_check();
+		ut_ad(!node->table->is_temporary());
+		err = row_undo_ins_remove_clust_rec(node);
 	}
 
 	dict_table_close(node->table, dict_locked, FALSE);
