@@ -2319,7 +2319,7 @@ int ha_sphinx::extra ( enum ha_extra_function op )
 }
 
 
-int ha_sphinx::write_row ( byte * )
+int ha_sphinx::write_row ( const byte * )
 {
 	SPH_ENTER_METHOD();
 	if ( !m_pShare || !m_pShare->m_bSphinxQL )
@@ -2749,7 +2749,9 @@ const Item * ha_sphinx::cond_push ( const Item *cond )
 		if ( !m_pShare->m_bSphinxQL )
 		{
 			// on non-QL tables, intercept query=value condition for SELECT
-			if (!( args[0]->type()==Item::FIELD_ITEM && args[1]->type()==Item::STRING_ITEM ))
+			if (!( args[0]->type()==Item::FIELD_ITEM &&
+			       args[1]->is_of_type(Item::CONST_ITEM,
+			                           STRING_RESULT)))
 				break;
 
 			Item_field * pField = (Item_field *) args[0];
@@ -2765,7 +2767,9 @@ const Item * ha_sphinx::cond_push ( const Item *cond )
 
 		} else
 		{
-			if (!( args[0]->type()==Item::FIELD_ITEM && args[1]->type()==Item::INT_ITEM ))
+			if (!( args[0]->type()==Item::FIELD_ITEM &&
+			       args[1]->is_of_type(Item::CONST_ITEM,
+			                           INT_RESULT)))
 				break;
 
 			// on QL tables, intercept id=value condition for DELETE
